#include <metal_stdlib>

using namespace metal;

constant float4x4 IDENTITY_MATRIX = float4x4(
    1, 0, 0, 0,
    0, 1, 0, 0,
    0, 0, 1, 0,
    0, 0, 0, 1);

constant float PI = 3.14159265358979323846264;

constexpr sampler SAMPLER_NEAREST (mag_filter::nearest,
                                   min_filter::nearest);

struct Uniforms
{
    float rotation;
    float aspectRatio;
    int randomseed;
    bool loading;
    bool isFreqCurve;
};

struct FragmentUniforms
{
    int coefsCount;
    float yOffset;
};

struct VertexIn
{
    float4 position;
    float4 color;
};

struct VertexOut
{
    float4 position [[position]];
    float4 color;
    float2 uv;
    int randomseed;
    bool loading;
};

float cantorPairing(float x, float y) {
    float z = (x + y) * (x + y + 1) / 2.0 + y;

    // Normalize the result to be between 0 and 1
    float maxVal = 2.0;
    return z / ((maxVal * (maxVal + 1)) / 2.0 + 1.0);
}

// https://developer.apple.com/library/archive/samplecode/MetalShaderShowcase/Listings/MetalShaderShowcase_AAPLWoodShader_metal.html
// Generate a random float in the range [0.0f, 1.0f] (based on the xor128 algorithm)
float rand(int seed) {
    seed = (seed<< 13) ^ seed;
    return 1.0 - ((seed*seed*seed*15731 + seed*789221 + 1376312589) & 2147483647) / 1073741824.0f / 2.0f;
}

float clamp01(float in) {
    return clamp(in, 0.0, 1.0);
}

float4x4 translate(float x, float y, float z) {
    return float4x4(
        1, 0, 0, 0,
        0, 1, 0, 0,
        0, 0, 1, 0,
        x, y, z, 1);
}

float4x4 rotate(float x, float y, float z) {
    float c, s;

    c = cos(x);
    s = sin(x);
    float4x4 rotX = float4x4(
        1, 0, 0, 0,
        0, c,-s, 0,
        0, s, c, 0,
        0, 0, 0, 1);

    c = cos(y);
    s = sin(y);
    float4x4 rotY = float4x4(
        c, 0, s, 0,
        0, 1, 0, 0,
       -s, 0, c, 0,
        0, 0, 0, 1);

    c = cos(z);
    s = sin(z);
    float4x4 rotZ = float4x4(
        c,-s, 0, 0,
        s, c, 0, 0,
        0, 0, 1, 0,
        0, 0, 0, 1);

    return rotZ * rotY * rotX;
}

float4x4 scale(float x, float y, float z)
{
    return float4x4(
        x, 0, 0, 0,
        0, y, 0, 0,
        0, 0, z, 0,
        0, 0, 0, 1);
}

float4x4 perspective(float fovY, float aspect, float nearZ, float farZ) {
    float yscale = 1 / tan(fovY * 0.5);
    float xscale = yscale / aspect;
    float q = farZ / (farZ - nearZ);

    return float4x4(
        xscale, 0, 0, 0,
        0, yscale, 0,  0,
        0, 0, q, 1,
        0, 0, -q * nearZ, 0);
}

vertex VertexOut vertexFunction(
    device VertexIn *vertices [[buffer(0)]],
    constant Uniforms &uniforms [[buffer(1)]],
    uint vid [[vertex_id]])
{
    float4x4 transform = IDENTITY_MATRIX;
    if(uniforms.isFreqCurve) {
        transform = transform * scale(2, 2, 1);
    } else {
        transform = transform * translate(0, 0, 2);
        transform = transform * rotate(0, uniforms.rotation, 0);
        transform = perspective(PI/4, uniforms.aspectRatio, 0.1, 100) * transform;
    }

    VertexOut out;
    out.position = transform * vertices[vid].position;
    out.color = vertices[vid].color;
    out.uv = float2(vertices[vid].position.x + 0.5, 1 - (vertices[vid].position.y + 0.5));
    out.randomseed = uniforms.randomseed;
    out.loading = uniforms.loading;
    return out;
}

fragment float4 fragmentFunction(
    VertexOut in [[stage_in]],
    texture2d<float> texture [[texture(0)]],
    constant FragmentUniforms &fragmentUniforms [[buffer(0)]],
    constant float *coefs [[buffer(1)]])
{
    if(fragmentUniforms.coefsCount != 0) {
        float y = 0;
        for(int i = 0; i < fragmentUniforms.coefsCount; i++) {
            float a0 = coefs[0 + i*5];
            float a1 = coefs[1 + i*5];
            float a2 = coefs[2 + i*5];
            float b1 = coefs[3 + i*5];
            float b2 = coefs[4 + i*5];
            float w = exp(log(1000.0)*in.uv.x)*PI/1000.0;
            float phi = pow(sin(w/2), 2);
            y += log(pow(a0+a1+a2, 2) - 4*(a0*a1 + 4*a0*a2 + a1*a2)*phi + 16*a0*a2*phi*phi) - log(pow(1+b1+b2, 2) - 4*(b1 + 4*b2 + b1*b2)*phi + 16*b2*phi*phi);
        }
        float dist = in.uv.y - fragmentUniforms.yOffset + y/16;
        if(dist > 0) {
            float scale = 20;
            float scaledY = y*10/log(10.0)/scale + 0.5;
            float alpha = clamp01(1 - dist*10);
            alpha = pow(alpha, 10);
            return alpha * float4(clamp01(scaledY*2), clamp01((1-scaledY)*2), 0, 1);
        } else {
            return float4(0, 0, 0, 0);
        }
    } else {
        float rgb = rand(in.randomseed + int(100000 * cantorPairing(in.uv.x, in.uv.y)));
        float4 noise = float4(rgb, rgb, rgb, 1);
        if(in.loading) {
            return noise;
        } else {
            return 0.5 * in.color + 0.5 * noise;
        }
    }
}
